/*
 * HARK/hark-base
 *
 * Copyright (C) 2008-2018
 *
 */

#if !defined(HARK_BASE_NODE_H_INCLUDED)
#define HARK_BASE_NODE_H_INCLUDED 1

#include "common.h"

#include "Object.h"
#include <string>
#include <map>
#include <vector>
#include "ObjectRef.h"
#include "BaseException.h"
#include <typeinfo>
#include "ParameterSet.h"
#include "NodeFactory.h"

namespace HARK
{
    using factoryEntry = std::map<std::string, _NodeFactory*>::value_type;

    class Node;

    class NodeInput
    {
    public:
	int outputID;
	Node *node;
	std::string name;

	NodeInput (const NodeInput &in)
	{
	    node = in.node;
	    outputID = in.outputID;
	    name = in.name;
	}

	NodeInput& operator= (const NodeInput &in)
	{
	    if (&in != this) {
		node = in.node;
		outputID = in.outputID;
		name = in.name;
	    }
	    return *this;
	}

	NodeInput() : outputID(-1), node(NULL) {} //-1 means unused

	NodeInput(const std::string &inputName) : outputID(-1), node(NULL), name(inputName) {}

    private:
    };

    class Node : public Object
    {
    protected:
	ParameterSet parameters;

	virtual void connectToNode(unsigned int in, Node *inputNode, unsigned int out);

	virtual int addOutput(const std::string &outputName);

	virtual int addInput(const std::string &inputName);

	virtual std::vector<NodeInput>& getInputs ();

    public:
	Node(const std::string &nodeName, const ParameterSet &params);

	virtual ~Node();

	virtual ObjectRef getOutput(int output_id, int count) = 0;

	virtual ObjectRef getOutputNamed (const std::string &outputName, int count);

	virtual void connectToNode(const std::string &in, Node *inputNode, const std::string &out);

	virtual void initialize ();

	virtual bool hasOutput(int output_id) const;

	ObjectRef getInput(int inputID, int count);

	std::string getName() const;

	virtual void request(int outputID, const ParameterSet &req);

	void printOn(std::ostream &out = std::cout) const;

	static int addFactory (const std::string &factoryName, _NodeFactory* const factory);
	static _NodeFactory* getFactoryNamed (const std::string &name);

    protected:
	Node();

    public:
	virtual int translateInput(const std::string &inputName);
	virtual int translateOutput(const std::string &outputName);
	static std::map<std::string, _NodeFactory*> &factoryDictionary();

	virtual int getInputID(const std::string &inputName);
	virtual int getOutputID(const std::string &inputName);

    protected:
	std::string name;

    public:
	std::vector<NodeInput> inputs;
	std::vector<std::string> outputNames;

    };

    class NotInitializedException : public BaseException
    {

    public:
	NotInitializedException (std::map<std::string,Node * > aMap)
	{
	    nodeMap = aMap;
	}

	virtual void print(std::ostream &out = std::cerr) const
	{
	    out<<"NotInitializedException occured"<<std::endl;

	    std::map<std::string,Node*>::const_iterator iter;

	    for (iter = nodeMap.begin(); iter != nodeMap.end(); iter++) {
		out<<"This node is not initialized: "<<(*iter).first<<std::endl;
	    }
	}

	std::map<std::string,Node*> nodeMap;
    };

    class NodeException : public BaseException
    {
    public:
	NodeException( Node *_node, std::string _message, const char *_file, int _line) :
	    message(_message),
	    node(_node),
	    file(_file),
	    line(_line),
	    frozen(false)
	{}

	virtual ~NodeException() {}

	virtual void print(std::ostream &out = std::cerr) const;
	virtual void freeze();

    protected:
	std::string message;
	Node *node;
	std::string file;
	int line;
	bool frozen;

    };

#define DECLARE_NODE_XPM(NodeTypeName, XPMFilePtr) static int dummy_initializer_for ## NodeTypeName = \
						       Node::addFactory (# NodeTypeName, new NodeFactory<NodeTypeName>(# NodeTypeName));

#define DECLARE_NODE(NodeTypeName) DECLARE_NODE_XPM(NodeTypeName,NULL)

}

#endif

/* end of file */
