/*
 * HARK/hark-base
 *
 * Copyright (C) 2008-2018
 *
 */

#if !defined(HARK_BASE_OPERATORS_H_INCLUDED)
#define HARK_BASE_OPERATORS_H_INCLUDED 1

#include "common.h"

#include "ObjectRef.h"
#include <map>
#include <unordered_map>
#include <sstream>

namespace HARK
{
    /* 抽象化された ObjectRef 間演算子 */
    ObjectRef operator +  (ObjectRef x, ObjectRef y);
    ObjectRef operator -  (ObjectRef x, ObjectRef y);
    ObjectRef operator *  (ObjectRef x, ObjectRef y);
    ObjectRef operator /  (ObjectRef x, ObjectRef y);
    ObjectRef operator <  (ObjectRef x, ObjectRef y);
    ObjectRef operator >  (ObjectRef x, ObjectRef y);
    ObjectRef operator == (ObjectRef x, ObjectRef y);
    ObjectRef concat(ObjectRef x, ObjectRef y);
    ObjectRef max(ObjectRef x, ObjectRef y);
    ObjectRef min(ObjectRef x, ObjectRef y);

    /* 新たに ObjectRef 間の演算を定義する場合、以下を参照する */
#if 0
    /* 以下は、例として Map<String, String> 同士の比較を行う演算子を独自に定義する場合について述べる */
    
    /* 1. x を左辺値、Y を右辺値として、以下のようなクラスを定義する
       クラスは、ObjectRef ２つを引数に持ち、ObjectRef を返す static メソッド Function をもつ。 */
    class MyFunctionClass
    {
	static ObjectRef Function(ObjectRef x, ObjectRef y)
	{
	    RCPtr<Map<String, String> > vx = x;
	    RCPtr<Map<String, String> > vy = y;
	    RCPtr<Bool> vz(Bool::alloc(1));
	    return vz;
	}
    };

    /* 2. 定義した関数を、テーブルに登録する。
       ここで、GetOperatorDispatchTable の引数には、OperatorDispatch で
       定義されている enum 値である下記のいずれかを与える。
       EQUAL, ADD, SUB, MUL, DIV, LESS, GREATER, MIN, MAX, CONCAT,
    */
    OperatorDispatch &operatorDispatch = OperatorDispatch::GetInstance();
    FunctorDispatch &functorDispatch = operatorDispatch.GetOperatorDispatchTable[OperatorDispatch::EQUAL];
    functorDispatch.addFunctor<Map<String, String>, Map<String, String>, Bool, Function>();

#endif

    using Functor = ObjectRef (*)(ObjectRef, ObjectRef);

    class FunctorDispatch
    {
    public:
	using key_type   = std::pair<TypeInfo, TypeInfo>;
	using value_type = Functor;
	//using table_type = std::map<key_type, value_type>;
	using table_type = std::unordered_map<key_type, value_type, TypeInfoPairHash>;

	table_type dispatchTable;
	
    public:
	template<typename X, typename Y, typename Z, typename W,
		 template<typename, typename, typename, typename, template<typename, typename, typename> typename> typename F,
		 template<typename, typename, typename> typename OP>
	void addFunctor()
	{
	    const std::type_info *tx = &typeid(X);
	    const std::type_info *ty = &typeid(Y);
	    std::pair<TypeInfo, TypeInfo> key(tx, ty);

	    //std::cout << "operators: addFunctor<X, Y, Z, W, OP, F>: type(x): " << key.first << ", type(y): " << key.second << ", functor: " << typeid(F<X, Y, Z, W, OP>).name() << ", table=" << (const void *)&this->dispatchTable << ", this=" << (const void *)this << std::endl;
	    
	    this->dispatchTable[key] = F<X, Y, Z, W, OP>::Function;
	}

	template<typename X, typename Y, typename Z, typename F>
	void addFunctor()
	{
	    const std::type_info *tx = &typeid(X);
	    const std::type_info *ty = &typeid(Y);
	    std::pair<TypeInfo, TypeInfo> key(tx, ty);

	    //std::cout << "operators: addFunctor<X, Y, Z, F>: type(x): " << key.first << ", type(y): " << key.second << ", functor: " << typeid(F).name() << ", table=" << (const void *)&this->dispatchTable << ", table=" << (const void *)&this->dispatchTable << ", this=" << (const void *)this << std::endl;
	    
	    this->dispatchTable[key] = F::Function;
	}
	
	ObjectRef
	call(ObjectRef x, ObjectRef y) const
	{
	    const std::type_info *tx = &typeid(*x);
	    const std::type_info *ty = &typeid(*y);
	    std::pair<TypeInfo, TypeInfo> key(tx, ty);

	    //std::cout << "operators: call: type(x): " << tx->name() << " (" << (const void *)tx << "), type(y): " << ty->name() << " (" << (const void *)ty << "), table=" << (const void *)&this->dispatchTable << ", this=" << (const void *)this << std::endl;

	    table_type::const_iterator it;
	    if((it = this->dispatchTable.find(key)) != this->dispatchTable.end()){
		return it->second(x, y);
	    }else{
		std::stringstream ss;
		ss << "Failed to dispatch operation between " << tx->name() << " and " << ty->name();
		throw new GeneralException(ss.str(), __FILE__, __LINE__);
	    }
	}
    };

    class OperatorDispatch
    {
    public:
	enum Operation {
	    EQUAL, ADD, SUB, MUL, DIV, LESS, GREATER, MIN, MAX, CONCAT,
	};

    public:
	OperatorDispatch();
	
	const FunctorDispatch &operator [] (enum Operation op)
	{
	    return *this->dispatchTable[op];
	}

    private:
	std::map<Operation, FunctorDispatch *> dispatchTable;

    };
}

#endif

/* end of file */
