/*
 * HARK/hark-base
 *
 * Copyright (C) 2008-2018
 *
 */

#include "Sleep.h"

#include "Node.h"
#include <unistd.h>
#include <csignal>
#include <time.h>
#include <sys/types.h>
#include <sys/timeb.h>

#ifdef _WIN32
#include <windows.h>
#endif

namespace HARK
{
    DECLARE_NODE(Sleep);
    /*Node
     *
     * @name Sleep
     * @category Flow
     * @description Sleep a certain amount of time.
     *
     * @output_name VALUE
     * @output_description Always return TRUE.
     *
     * @parameter_name SECONDS
     * @parameter_description Sleep x seconds.
     * @parameter_type float
     *
     END*/

    static bool _stop_flag = false;
    
    static
    void
    signal_handler(int s)
    {
	if(s = SIGINT) {
	    _stop_flag = true;
	}
    }

    Sleep::Sleep(const std::string &nodeName, const ParameterSet &params) :
        Node(nodeName, params)
    {
        this->_time = dereference_cast<float>(parameters.get("SECONDS"));
        this->_mSec = this->_time*1000;

        this->_outputID = this->addOutput("VALUE");
        signal(SIGINT, signal_handler);
    }

    int
    Sleep::getMilliCount() const
    {
        timeb tb;
        ftime(&tb);
        int nCount = tb.millitm + (tb.time & 0xfffff) * 1000;
        return nCount;
    }

    int
    Sleep::getMilliSpan(int nTimeStart) const
    {
        int nSpan = this->getMilliCount() - nTimeStart;
        if(nSpan < 0) {
            nSpan += 0x100000 * 1000;
        }
        return nSpan;
    }

    ObjectRef
    Sleep::getOutput(int output_id, int count)
    {
        this->_pt = this->getMilliCount();

        while(1) {
            if(_stop_flag) {
		std::cout << std::endl << "User stop" << std::endl;

                pid_t pid=getpid();
#ifdef _WIN32
                HANDLE hProcess;
                hProcess = OpenProcess(PROCESS_TERMINATE, FALSE, pid);
                if (hProcess == NULL) {
                    CloseHandle(hProcess);
                    break;
                }
                if (!TerminateProcess(hProcess, (unsigned int)-1)) {
                    CloseHandle(hProcess);
                    break;
                }
                CloseHandle(hProcess);
#else
                kill(pid, SIGTERM);
#endif

            }
            if(this->getMilliSpan(this->_pt) > this->_mSec) {
                break;
            }
#ifdef _WIN32
            _sleep(1);
#else
            usleep(1000);
#endif
        }

        if (output_id == this->_outputID) {
            return TrueObject;
        } else {
            throw new NodeException (this, "Sleep: Unknown output id", __FILE__, __LINE__);
        }
    }

}

/* end of file */
