/*
 * HARK/hark-base
 *
 * Copyright (C) 2008-2018
 *
 */

#include "ObjectRef.h"
#include "conversion.h"
#include "Matrix.h"
#include "Vector.h"

#include <sstream>

namespace HARK
{
    template<class X, class Y>
    struct ScalarScalarFunctor
    {
	static ObjectRef Function(ObjectRef x)
	{
	    RCPtr<X> vx = x;
	    RCPtr<Y> vy(Y::alloc(static_cast<typename Y::basicType>(vx->val())));
	    return vy;
	}
    };

    template<class X, class Y>
    struct ScalarVectorFunctor
    {
	static ObjectRef Function(ObjectRef x)
	{
	    RCPtr<X> vx = x;
	    RCPtr<Y> vy(Y::alloc(1));
	    (*vy)[0] = static_cast<typename Y::basicType>(vx->val());
	    return vy;
	}
    };

    template<class X, class Y>
    struct ScalarMatrixFunctor
    {
	static ObjectRef Function(ObjectRef x)
	{
	    RCPtr<X> vx = x;
	    RCPtr<Y> vy(new Y(1, 1));
	    (*vy)(0, 0) = static_cast<typename Y::basicType>(vx->val());
	    return vy;
	}
    };

    template<class X, class Y>
    struct VectorVectorFunctor
    {
	static ObjectRef Function(ObjectRef x)
	{
	    RCPtr<X> vx = x;
	    RCPtr<Y> vy(Y::alloc(vx->size()));
	    for(int i = 0; i < vx->size(); i++){
		(*vy)[i] = static_cast<typename Y::basicType>((*vx)[i]);
	    }
	    return vy;
	}
    };

    template<class X, class Y>
    struct VectorMatrixFunctor
    {
	static ObjectRef Function(ObjectRef x)
	{
	    RCPtr<X> vx = x;
	    RCPtr<Y> vy(new Y(1, vx->size()));
	    for(int i = 0; i < vx->size(); i++){
		(*vy)(0, i) = static_cast<typename Y::basicType>((*vx)[i]);
	    }
	    return vy;
	}
    };

    template<class X, class Y>
    struct MatrixMatrixFunctor
    {
	static ObjectRef Function(ObjectRef x)
	{
	    RCPtr<X> vx = x;
	    RCPtr<Y> vy(new Y(vx->nrows(), vx->ncols()));
	    for(int i = 0; i < vx->nrows(); i++){
		for(int j = 0; j < vx->ncols(); j++){
		    (*vy)(i, j) = static_cast<typename Y::basicType>((*vx)(i, j));
		}
	    }
	    return vy;
	}
    };

    template<class X, class Y>
    struct ScalarStringFunctor
    {
	static ObjectRef Function(ObjectRef x)
	{
	    std::ostringstream os;
	    x->prettyPrint(os);
	    RCPtr<Y> vy(new Y(os.str()));
	    return vy;
	}
    };

    template<class X, class Y>
    struct StringScalarFunctor
    {
	static ObjectRef Function(ObjectRef x)
	{
	    RCPtr<X> vx = x;
	    std::istringstream is(*vx);

	    typename Y::basicType v;
	    is >> v;
	    RCPtr<Y> vy(new Y(v));
	    return vy;
	}
    };

    template<class X, class Y>
    struct NilFunctor
    {
	static ObjectRef Function(ObjectRef x)
	{
	    return nilObject;
	}
    };

    ConversionDispatch ConversionDispatch::theInstance;
    
    ConversionDispatch &
    ConversionDispatch::GetInstance()
    {
	return ConversionDispatch::theInstance;
    }

    ConversionDispatch::ConversionDispatch()
    {
	this->AddConversionFunc<Bool, Bool, ScalarScalarFunctor>();
	this->AddConversionFunc<Int, Bool, ScalarScalarFunctor>();
	this->AddConversionFunc<Float, Bool, ScalarScalarFunctor>();
	this->AddConversionFunc<Double, Bool, ScalarScalarFunctor>();
	this->AddConversionFunc<Bool, Int, ScalarScalarFunctor>();
	this->AddConversionFunc<Int, Int, ScalarScalarFunctor>();
	this->AddConversionFunc<Float, Int, ScalarScalarFunctor>();
	this->AddConversionFunc<Double, Int, ScalarScalarFunctor>();
	this->AddConversionFunc<Bool, Float, ScalarScalarFunctor>();
	this->AddConversionFunc<Int, Float, ScalarScalarFunctor>();
	this->AddConversionFunc<Float, Float, ScalarScalarFunctor>();
	this->AddConversionFunc<Double, Float, ScalarScalarFunctor>();
	this->AddConversionFunc<Bool, Double, ScalarScalarFunctor>();
	this->AddConversionFunc<Int, Double, ScalarScalarFunctor>();
	this->AddConversionFunc<Float, Double, ScalarScalarFunctor>();
	this->AddConversionFunc<Double, Double, ScalarScalarFunctor>();
	this->AddConversionFunc<Bool, Complex<float>, ScalarScalarFunctor>();
	this->AddConversionFunc<Int, Complex<float>, ScalarScalarFunctor>();
	this->AddConversionFunc<Float, Complex<float>, ScalarScalarFunctor>();
	this->AddConversionFunc<Double, Complex<float>, ScalarScalarFunctor>();
	this->AddConversionFunc<Complex<float>, Complex<float>, ScalarScalarFunctor>();
	this->AddConversionFunc<Complex<double>, Complex<float>, ScalarScalarFunctor>();
	this->AddConversionFunc<Bool, Complex<double>, ScalarScalarFunctor>();
	this->AddConversionFunc<Int, Complex<double>, ScalarScalarFunctor>();
	this->AddConversionFunc<Float, Complex<double>, ScalarScalarFunctor>();
	this->AddConversionFunc<Double, Complex<double>, ScalarScalarFunctor>();
	this->AddConversionFunc<Complex<float>, Complex<double>, ScalarScalarFunctor>();
	this->AddConversionFunc<Complex<double>, Complex<double>, ScalarScalarFunctor>();

	this->AddConversionFunc<Bool, Vector<bool>, ScalarVectorFunctor>();
	this->AddConversionFunc<Int, Vector<int>, ScalarVectorFunctor>();
	this->AddConversionFunc<Float, Vector<float>, ScalarVectorFunctor>();
	this->AddConversionFunc<Double, Vector<double>, ScalarVectorFunctor>();
	this->AddConversionFunc<Complex<float>, Vector<std::complex<float> >, ScalarVectorFunctor>();
	this->AddConversionFunc<Complex<double>, Vector<std::complex<double> >, ScalarVectorFunctor>();

	this->AddConversionFunc<Bool, Matrix<bool>, ScalarMatrixFunctor>();
	this->AddConversionFunc<Int, Matrix<int>, ScalarMatrixFunctor>();
	this->AddConversionFunc<Float, Matrix<float>, ScalarMatrixFunctor>();
	this->AddConversionFunc<Double, Matrix<double>, ScalarMatrixFunctor>();
	this->AddConversionFunc<Complex<float>, Matrix<std::complex<float> >, ScalarMatrixFunctor>();
	this->AddConversionFunc<Complex<double>, Matrix<std::complex<double> >, ScalarMatrixFunctor>();
	this->AddConversionFunc<Float, Matrix<std::complex<float> >, ScalarMatrixFunctor>();
	this->AddConversionFunc<Float, Matrix<std::complex<double> >, ScalarMatrixFunctor>();
	this->AddConversionFunc<Double, Matrix<std::complex<double> >, ScalarMatrixFunctor>();

	this->AddConversionFunc<Vector<int>, Vector<int>, VectorVectorFunctor>();
	this->AddConversionFunc<Vector<int>, Vector<float>, VectorVectorFunctor>();
	this->AddConversionFunc<Vector<int>, Vector<double>, VectorVectorFunctor>();
	this->AddConversionFunc<Vector<int>, Vector<std::complex<float> >, VectorVectorFunctor>();
	this->AddConversionFunc<Vector<int>, Vector<std::complex<double> >, VectorVectorFunctor>();
	this->AddConversionFunc<Vector<float>, Vector<int>, VectorVectorFunctor>();
	this->AddConversionFunc<Vector<float>, Vector<float>, VectorVectorFunctor>();
	this->AddConversionFunc<Vector<float>, Vector<double>, VectorVectorFunctor>();
	this->AddConversionFunc<Vector<float>, Vector<std::complex<float> >, VectorVectorFunctor>();
	this->AddConversionFunc<Vector<float>, Vector<std::complex<double> >, VectorVectorFunctor>();
	this->AddConversionFunc<Vector<double>, Vector<int>, VectorVectorFunctor>();
	this->AddConversionFunc<Vector<double>, Vector<float>, VectorVectorFunctor>();
	this->AddConversionFunc<Vector<double>, Vector<double>, VectorVectorFunctor>();
	this->AddConversionFunc<Vector<double>, Vector<std::complex<float> >, VectorVectorFunctor>();
	this->AddConversionFunc<Vector<double>, Vector<std::complex<double> >, VectorVectorFunctor>();
	this->AddConversionFunc<Vector<std::complex<float> >, Vector<std::complex<float> >, VectorVectorFunctor>();
	this->AddConversionFunc<Vector<std::complex<float> >, Vector<std::complex<double> >, VectorVectorFunctor>();
	this->AddConversionFunc<Vector<std::complex<double> >, Vector<std::complex<double> >, VectorVectorFunctor>();
	
	this->AddConversionFunc<Vector<int>, Matrix<int>, VectorMatrixFunctor>();
	this->AddConversionFunc<Vector<int>, Matrix<float>, VectorMatrixFunctor>();
	this->AddConversionFunc<Vector<int>, Matrix<double>, VectorMatrixFunctor>();
	this->AddConversionFunc<Vector<int>, Matrix<std::complex<float> >, VectorMatrixFunctor>();
	this->AddConversionFunc<Vector<int>, Matrix<std::complex<double> >, VectorMatrixFunctor>();
	this->AddConversionFunc<Vector<float>, Matrix<int>, VectorMatrixFunctor>();
	this->AddConversionFunc<Vector<float>, Matrix<float>, VectorMatrixFunctor>();
	this->AddConversionFunc<Vector<float>, Matrix<double>, VectorMatrixFunctor>();
	this->AddConversionFunc<Vector<float>, Matrix<std::complex<float> >, VectorMatrixFunctor>();
	this->AddConversionFunc<Vector<float>, Matrix<std::complex<double> >, VectorMatrixFunctor>();
	this->AddConversionFunc<Vector<double>, Matrix<int>, VectorMatrixFunctor>();
	this->AddConversionFunc<Vector<double>, Matrix<float>, VectorMatrixFunctor>();
	this->AddConversionFunc<Vector<double>, Matrix<double>, VectorMatrixFunctor>();
	this->AddConversionFunc<Vector<double>, Matrix<std::complex<float> >, VectorMatrixFunctor>();
	this->AddConversionFunc<Vector<double>, Matrix<std::complex<double> >, VectorMatrixFunctor>();
	this->AddConversionFunc<Vector<std::complex<float> >, Matrix<std::complex<float> >, VectorMatrixFunctor>();
	this->AddConversionFunc<Vector<std::complex<float> >, Matrix<std::complex<double> >, VectorMatrixFunctor>();
	this->AddConversionFunc<Vector<std::complex<double> >, Matrix<std::complex<double> >, VectorMatrixFunctor>();

	this->AddConversionFunc<Matrix<int>, Matrix<int>, MatrixMatrixFunctor>();
	this->AddConversionFunc<Matrix<int>, Matrix<float>, MatrixMatrixFunctor>();
	this->AddConversionFunc<Matrix<int>, Matrix<double>, MatrixMatrixFunctor>();
	this->AddConversionFunc<Matrix<int>, Matrix<std::complex<float> >, MatrixMatrixFunctor>();
	this->AddConversionFunc<Matrix<int>, Matrix<std::complex<double> >, MatrixMatrixFunctor>();
	this->AddConversionFunc<Matrix<float>, Matrix<int>, MatrixMatrixFunctor>();
	this->AddConversionFunc<Matrix<float>, Matrix<float>, MatrixMatrixFunctor>();
	this->AddConversionFunc<Matrix<float>, Matrix<double>, MatrixMatrixFunctor>();
	this->AddConversionFunc<Matrix<float>, Matrix<std::complex<float> >, MatrixMatrixFunctor>();
	this->AddConversionFunc<Matrix<float>, Matrix<std::complex<double> >, MatrixMatrixFunctor>();
	this->AddConversionFunc<Matrix<double>, Matrix<int>, MatrixMatrixFunctor>();
	this->AddConversionFunc<Matrix<double>, Matrix<float>, MatrixMatrixFunctor>();
	this->AddConversionFunc<Matrix<double>, Matrix<double>, MatrixMatrixFunctor>();
	this->AddConversionFunc<Matrix<double>, Matrix<std::complex<float> >, MatrixMatrixFunctor>();
	this->AddConversionFunc<Matrix<double>, Matrix<std::complex<double> >, MatrixMatrixFunctor>();
	this->AddConversionFunc<Matrix<std::complex<float> >, Matrix<std::complex<float> >, MatrixMatrixFunctor>();
	this->AddConversionFunc<Matrix<std::complex<float> >, Matrix<std::complex<double> >, MatrixMatrixFunctor>();
	this->AddConversionFunc<Matrix<std::complex<double> >, Matrix<std::complex<double> >, MatrixMatrixFunctor>();
	
	this->AddConversionFunc<Bool, String, ScalarStringFunctor>();
	this->AddConversionFunc<Int, String, ScalarStringFunctor>();
	this->AddConversionFunc<Float, String, ScalarStringFunctor>();
	this->AddConversionFunc<Double, String, ScalarStringFunctor>();
	this->AddConversionFunc<Complex<float>, String, ScalarStringFunctor>();
	this->AddConversionFunc<Complex<double>, String, ScalarStringFunctor>();

	this->AddConversionFunc<String, Bool, StringScalarFunctor>();
	this->AddConversionFunc<String, Int, StringScalarFunctor>();
	this->AddConversionFunc<String, Float, StringScalarFunctor>();
	this->AddConversionFunc<String, Double, StringScalarFunctor>();
    }

    ObjectRef
    ConversionDispatch::call(const std::type_info *tx, const std::type_info *ty, ObjectRef x)
    {
	std::pair<TypeInfo, TypeInfo> key(tx, ty);

	table_type::const_iterator it;
	if((it = this->dispatchTable.find(key)) == this->dispatchTable.end()){
	    std::stringstream ss;
	    ss << "Failed to dispatch conversion between " << tx->name() << " and " << ty->name();
	    throw new GeneralException(ss.str(), __FILE__, __LINE__);
	}else{
	    return it->second(x);
	}
    }

    template<typename X, typename Y, template<typename, typename> typename F>
    void
    ConversionDispatch::AddConversionFunc()
    {
	key_type key(&typeid(X), &typeid(Y));
	this->dispatchTable[key] = F<X, Y>::Function;
    }
}

/* end of files */


